from pathlib import Path
from typing import List, Optional, Union

import pandas as pd


def create_excel_summary_report(
    df: pd.DataFrame,
    destination_path: Union[str, Path],
    group_columns: List[str],
    *,
    individual_sheet_name: str = "Individual",
    average_sheet_name: str = "Average",
    std_sheet_name: str = "Std",
    excel_writer: Optional[pd.ExcelWriter] = None,
):
    """Creates a summary report of the given DataFrame and saves it to an Excel file.
    It groups the rows using the given `group_columns` and calculates the mean and
    standard deviation of the numeric columns for each group.

    Parameters
    ----------
    df : pd.DataFrame
        The DataFrame to create the report from.
    destination_path : Union[str, Path]
        The path to save the Excel file. It is ignored if `excel_writer` is provided.
    group_columns : List[str]
        The columns to group the rows by.
    individual_sheet_name : str, optional, default="Individual"
        The name of the sheet for the individual rows.
    average_sheet_name : str, optional, default="Average"
        The name of the sheet for the mean values.
    std_sheet_name : str, optional, default="Std"
        The name of the sheet for the standard deviation values.
    excel_writer : Optional[pd.ExcelWriter], optional, default=None
        A pd.ExcelWriter object that will be used to write the dataframes to an excel
        file. If None, a new Excel file will be created, by default None. Using an
        external ExcelWriter can be useful if you want to write additional sheets to
        the report excel file.

    Returns
    -------
    Path
        The path to the created Excel file.
    """

    destination_path = Path(destination_path)

    mean_df = df.groupby(group_columns).mean(numeric_only=True).reset_index()
    std_df = df.groupby(group_columns).std(numeric_only=True).reset_index()

    destination_path.parent.mkdir(parents=True, exist_ok=True)

    if excel_writer is None:
        with pd.ExcelWriter(destination_path) as writer:
            df.to_excel(writer, sheet_name=individual_sheet_name, index=False)
            mean_df.to_excel(writer, sheet_name=average_sheet_name, index=False)
            std_df.to_excel(writer, sheet_name=std_sheet_name, index=False)
    else:
        df.to_excel(excel_writer, sheet_name=individual_sheet_name, index=False)
        mean_df.to_excel(excel_writer, sheet_name=average_sheet_name, index=False)
        std_df.to_excel(excel_writer, sheet_name=std_sheet_name, index=False)

    return destination_path


def create_excel_columns_report(
    df: pd.DataFrame,
    destination_path: Union[str, Path],
    metric_columns: List[str],
    pivot_index: str,
    pivot_columns: List[str],
    *,
    excel_writer: Optional[pd.ExcelWriter] = None,
):
    """Create an Excel report with multiple sheets, each containing a pivot table of the
    given DataFrame. The pivot table is created by pivoting the DataFrame using the
    given `pivot_index` and `pivot_columns`. The values of the pivot table are the
    columns specified in `metric_columns`. Each sheet in the Excel file corresponds to
    a column in `metric_columns`.

    To create an Excel file that contains one sheet for each metric, one column for each
    combination of methodology and dataset, and one row for each seed, the metric_columns
    should be the list of metric columns, the pivot_index should be the seed column and
    the pivot_columns should be the methodology and dataset columns.

    Parameters
    ----------
    df : pd.DataFrame
        The DataFrame to create the report from.
    destination_path : Union[str, Path]
        The path to save the Excel file. It is ignored if `excel_writer` is provided.
    metric_columns : List[str]
        The columns to create pivot tables for.
    pivot_index : str
        The column to use as the index of the pivot table.
    pivot_columns : List[str]
        The columns to use as the columns of the pivot table.
    excel_writer : Optional[pd.ExcelWriter], optional, default=None
        A pd.ExcelWriter object that will be used to write the dataframes to an excel
        file. If None, a new Excel file will be created, by default None. Using an
        external ExcelWriter can be useful if you want to write additional sheets to
        the report excel file.

    Returns
    -------
    Path
        The path to the created Excel file.
    """

    destination_path = Path(destination_path)

    destination_path.parent.mkdir(parents=True, exist_ok=True)

    def _create_pivot_df(column):
        pivot_df = (
            df.pivot(
                index=pivot_index,
                columns=pivot_columns,
                values=column,
            )
            .reset_index()
            .drop(columns=pivot_index, level=0)
        )
        pivot_df.columns = pivot_df.columns.map("_".join)

        return pivot_df

    if excel_writer is None:
        with pd.ExcelWriter(destination_path) as writer:
            for column in metric_columns:
                pivot_df = _create_pivot_df(column)

                pivot_df.to_excel(writer, sheet_name=column, index=False)
    else:
        for column in metric_columns:
            pivot_df = _create_pivot_df(column)

            pivot_df.to_excel(excel_writer, sheet_name=column, index=False)

    return destination_path



import copy

import numpy as np


class NonDefaultStrMethodError(Exception):
    """Raised when an object does not have a custom __str__ method."""

    pass


def sanitize_json(data, accept_default_str=True):
    """Returns an object that can be safely serialized to json.

    Numpy arrays are converted to lists.
    Lists, tuples and dictionary are processed recursively.
    When no conversion is possible, the object is converted to a string.

    Parameters
    ----------
    data : Any
        The data to be sanitized.

    accept_default_str : bool, optional
        If True, the function will use the default string representation of the object
        even if it does not have a custom implementation of the __str__ method.
        If False, the function will raise an error if the object does not have a custom
        implementation of the __str__ method.
    Returns
    -------
    sanitized_data : Any
        The sanitized data.

    Raises
    ------
    ValueError
        If the data cannot be sanitized.
    """

    data = copy.deepcopy(data)

    if data is None:
        return None
    elif isinstance(data, (int, float, str, bool, str)):
        return data
    elif isinstance(data, np.ndarray):
        return data.tolist()
    elif isinstance(data, (list, tuple)):
        return [sanitize_json(d, accept_default_str) for d in data]
    elif isinstance(data, dict):
        return {k: sanitize_json(v, accept_default_str) for k, v in data.items()}
    elif hasattr(data, "__str__"):
        if accept_default_str or type(data).__str__ is not object.__str__:
            return str(data)
        else:
            raise NonDefaultStrMethodError(
                f"Object of type {type(data)} does not have a custom __str__ method."
                " Set accept_default_str to True to use the default __str__ method or"
                " implement a custom __str__ method."
            )
    else:
        raise ValueError(f"Cannot serialize data of type {type(data)} to json.")



from .array import check_array
from .dicts import dict_contains_dict, get_deep_item_from_dict
from .json import NonDefaultStrMethodError, sanitize_json

__all__ = [
    "sanitize_json",
    "dict_contains_dict",
    "NonDefaultStrMethodError",
    "get_deep_item_from_dict",
    "check_array",
]



import warnings
from typing import Callable, Dict, List, Literal, Optional, Union

import numpy as np

from ..result import Result
from ..utils import get_deep_item_from_dict


def get_metric_columns_values(
    targets: Optional[Union[np.ndarray, list]],
    predictions: Optional[Union[np.ndarray, list]],
    prefix: str,
    metrics_fn: Callable[[np.ndarray, np.ndarray], Dict[str, float]],
    raise_errors: Literal["error", "warning", "ignore"] = "error",
) -> Dict[str, float]:
    """Creates the row with the metrics values for the given targets and predictions.
    The name of each column is determined by appending the name of the metric to the
    given prefix. For example, if adding the training metrics, the prefix could be
    'train_', so that the columns are named 'train_<metric_name>'.

    Parameters
    ----------
    targets : np.ndarray or list
        The targets.
    predictions : np.ndarray or list
        The predictions.
    prefix : str
        The prefix to add to the column names.
    metrics_fn : Callable[[np.ndarray, np.ndarray], Dict[str, float]]
        The function to calculate the metrics. See `ResultSet.get_dataframe` for more
        details.
    raise_errors : Literal["error", "warning", "ignore"], default="error"
        If set to 'error', it will raise an error if the targets or predictions are not
        valid values. If set to 'warning', it will throw a warning instead. If set to
        'ignore', it will ignore the error and return an empty dictionary.

    Returns
    -------
    row : Dict[str, float]
        The row with the metrics values. The keys represent the name of the columns and
        the values are the metrics values.

    Raises
    ------
    ValueError
        If raise_errors is not 'error', 'warning' or 'ignore'.
    TypeError
        If the targets or predictions are not numpy arrays and raise_errors is 'error'.
    TypeError
        If the metrics function does not return a dictionary.
    """

    if raise_errors not in ["error", "warning", "ignore"]:
        raise ValueError(
            f"raise_errors must be 'error', 'warning' or 'ignore'."
            f" Found: {raise_errors}."
        )

    if isinstance(targets, list):
        targets = np.array(targets)

    if isinstance(predictions, list):
        predictions = np.array(predictions)

    if targets is not None and not isinstance(targets, np.ndarray):
        if raise_errors == "error":
            raise TypeError(
                f"If set, targets must be a numpy array. Found type: {type(targets)}."
                " You can set raise_errors='warning' to skip this error and throw a "
                "warning instead. You can also set raise_errors='ignore' to ignore "
                "this error."
            )
        else:
            targets = np.array([])
            if raise_errors == "warning":
                warnings.warn(
                    f"If set, targets must be a numpy array. Found type: {type(targets)}."
                    " Setting targets to an empty numpy array."
                )

    if predictions is not None and not isinstance(predictions, np.ndarray):
        if raise_errors == "error":
            raise TypeError(
                f"If set, predictions must be a numpy array."
                f" Found type: {type(predictions)}."
                " You can set raise_errors='warning' to skip this error and throw a "
                "warning instead. You can also set raise_errors='ignore' to ignore "
                "this error."
            )
        else:
            predictions = np.array([])
            if raise_errors == "warning":
                warnings.warn(
                    f"If set, predictions must be a numpy array."
                    f" Found type: {type(predictions)}."
                    " Setting predictions to an empty numpy array."
                )

    if (
        targets is not None
        and predictions is not None
        and len(targets.shape) > 0
        and len(predictions.shape) > 0
        and targets.shape[0] > 0
        and predictions.shape[0] > 0
    ):
        metrics = metrics_fn(targets, predictions)
        if not isinstance(metrics, dict):
            raise TypeError(
                f"metrics_fn must return a dictionary. Found type: {type(metrics)}."
            )
        row = {f"{prefix}{column}": value for column, value in metrics.items()}
    else:
        if raise_errors in ["error", "warning"]:
            warnings.warn(
                "Targets or predictions are empty. Skipping metrics calculation."
                f" Prefix: {prefix}"
            )
        row = {}

    return row


def get_row_from_result(
    result: Result,
    config_columns: List[str] = [],
    metrics_fn=lambda targets, predictions: {},
    include_train: bool = False,
    include_val: bool = False,
    best_params_columns: List[str] = [],
    config_columns_prefix: str = "config_",
    best_params_columns_prefix: str = "best_",
    raise_errors: Literal["error", "warning", "ignore"] = "error",
) -> Dict[str, float]:
    """Create a row with the information of a `Result`, which can be included in the
    pandas DataFrame of a `ResultSet`. The row contains the configuration columns
    provided, the best parameters columns provided, the test metrics, and optionally the
    training and validation metrics.

    A list containing several dictionaries returned by this function, can be used to
    create a pandas DataFrame with the information of several `Result` objects.

    Parameters
    ----------
    result : Result
        The `Result` object.
    config_columns : List[str], default=[]
        The names of the columns to include from the configuration.
    metrics_fn : Callable[[np.ndarray, np.ndarray], Dict[str, float]],
                default=lambda targets, predictions: {}
        See `ResultSet.get_dataframe` for more details.
    include_train : bool, default=False
        Whether to include the training metrics.
    include_val : bool, default=False
        Whether to include the validation metrics.
    best_params_columns : List[str], default=[]
        The names of the columns to include from the best parameters.
    config_columns_prefix : str, default="config_"
        The prefix to add to the configuration columns.
    best_params_columns_prefix : str, default="best_"
        The prefix to add to the best parameters columns.
    raise_errors : Literal["error", "warning", "ignore"], default="error"
        Determines the behaviour when an error occurs during the calculation of the
        metrics. See `get_metric_columns_values` for more details.

    Returns
    -------
    row : Dict[str, float]
        The row with the information of the `Result`. Each key represents the name of a
        column and the value is the corresponding value.

    Raises
    ------
    TypeError
        If the input is not a `Result` object.
    """

    if not isinstance(result, Result):
        raise TypeError(f"result must be a Result object. Found type: {type(result)}.")

    data = result.get_data()

    if not config_columns_prefix:
        config_columns_prefix = ""

    if not best_params_columns_prefix:
        best_params_columns_prefix = ""

    targets = data.targets
    predictions = data.predictions
    time = data.time

    test_metrics = metrics_fn(targets, predictions)

    # Create row dict with config columns, best params columns and test metrics
    row = {
        **{
            f"{config_columns_prefix}{column}": get_deep_item_from_dict(
                result.config, column
            )
            for column in config_columns
        },
        **{
            f"{best_params_columns_prefix}{column}": get_deep_item_from_dict(
                data.best_params, column
            )
            for column in best_params_columns
        },
        **test_metrics,
    }

    if include_train:
        train_metrics_row = get_metric_columns_values(
            data.train_targets,
            data.train_predictions,
            "train_",
            metrics_fn,
            raise_errors=raise_errors,
        )
        row = {**row, **train_metrics_row}

    if include_val:
        val_metrics_row = get_metric_columns_values(
            data.val_targets,
            data.val_predictions,
            "val_",
            metrics_fn,
            raise_errors=raise_errors,
        )
        row = {**row, **val_metrics_row}

    row["time"] = time

    # Add best epoch and loss value if histories are available
    if (
        isinstance(data.train_history, np.ndarray)
        and len(data.train_history.shape) > 0
        and len(data.train_history) > 0
    ):
        row["best_train_epoch"] = data.train_history.argmin() + 1
        row["best_train_loss"] = data.train_history.min()

    if (
        isinstance(data.val_history, np.ndarray)
        and len(data.val_history.shape) > 0
        and len(data.val_history) > 0
    ):
        row["best_val_epoch"] = data.val_history.argmin() + 1
        row["best_val_loss"] = data.val_history.min()

    return row



from typing import List, Union

import numpy as np


def check_array(array: Union[np.ndarray, List], allow_none=False) -> np.ndarray:
    """Check if the input is a numpy array. If not, convert it to a numpy array.

    Parameters
    ----------
    array : Union[np.ndarray, List]
        The input array to check.
    allow_none : bool, optional
        Whether to allow the input to be None, by default False

    Returns
    -------
    np.ndarray
        The input array as a numpy array.
    """

    if array is None:
        if allow_none:
            return None
        else:
            raise TypeError("None is not a valid array")
    if not isinstance(array, np.ndarray):
        if isinstance(array, list):
            return np.array(array)
        else:
            raise TypeError(f"{array} must be a numpy array")
    return array



import importlib.util
import warnings
from pathlib import Path
from typing import Callable, Dict, List, Literal, Set, Union

import numpy as np
import pandas as pd

from ..result import Result
from ..utils import dict_contains_dict, sanitize_json
from .utils import get_row_from_result


class ResultSet:
    """Stores a set of `Result` objects.
    For each `Result` object, it stores the metadata of the experiment, such as the
    experiment info and the path where the result is stored. The predictions and targets
    are not loaded until needed.

    Attributes
    ----------
    results_ : Dict[str, Result]
        Dictionary that contains the config of the experiment as the key and the
        `Result` object as the value.
    """

    results_: Dict[str, Result]

    def __init__(self, results: Union[List[Result], Set[Result], Dict[str, Result]]):
        """Creates a `ResultSet` object.
        It can be initialised from a list, a set or a dictionary:
        - If a list or a set is provided, the config of the `Result` objects will be used
          as the key in the dictionary. If there are repeated configs, only the last
          `Result` object with that config will be stored.
        - If a dictionary is provided, it is assigned to the `results_` attribute.

        Parameters
        ----------
        results : Union[List[Result], Set[Result], Dict[str, Result]]
            The `Result` objects to store in the `ResultSet`.

        Raises
        ------
        TypeError
            If the `results` parameter is not a list, a set or a dictionary.
        """

        if isinstance(results, dict):
            self.results_ = results
        elif isinstance(results, (list, set)):
            self.results_ = {}
            for result in results:
                if not isinstance(result, Result):
                    raise TypeError(
                        f"Expected Result, got {type(result).__name__}",
                    )
                self.results_[str(sanitize_json(result.config))] = result
        else:
            raise TypeError(
                f"Expected list, set or dict, got {type(results).__name__}",
            )

    def contains(self, result: Union[str, dict, Result]) -> bool:
        """Checks if the `ResultSet` contains the given `Result`.

        Parameters
        ----------
        result : Union[str, dict, Result]
            The result that is searched in the `ResultSet`. It can be:
            - a string: the config dict of the result transformed to string,
            - a dict: the config dict of the result,
            - a `Result` object.

        If the result is a string, it is employed as the key to search in the
        `ResultSet`. If the result is a dict, it is json sanitized and transformed to
        string to be used as the key. If the result is a `Result` object, its config is
        sanitized and transformed to string to be used as the key.

        Returns
        -------
        contains : bool
            Whether the `ResultSet` contains the given result.

        Raises
        ------
        TypeError
            If the `result` parameter is not a str, a dict or a `Result`.
        """

        if isinstance(result, str):
            return result in self.results_
        elif isinstance(result, dict):
            return str(sanitize_json(result)) in self.results_
        elif isinstance(result, Result):
            return str(sanitize_json(result.config)) in self.results_
        else:
            raise TypeError(
                f"Expected str, dict or Result, got {type(result).__name__}",
            )

    def filter_by_config(self, config: dict) -> "ResultSet":
        """Create a copy of this `ResultSet` filtered by config. The new `ResultSet`
        contains the results that match the given config.

        Parameters
        ----------
        config : dict
            The config fields to filter by. To add a result to the filtered set, the
            result's config must contain all the fields in the config parameter with the
            same values. For example, if config={"a": 1, "b": 2}, the result config must
            contain both fields with those values. However, the result config can contain
            additional fields not listed in the provided config dictionary.
        Returns
        -------
        results : ResultSet
            A `ResultSet` that contains only the results that match the given config.

        Raises
        ------
        TypeError
            If the `config` parameter is not a dict.
        """

        if not isinstance(config, dict):
            raise TypeError(
                f"Expected dict, got {type(config).__name__}",
            )

        safe_config = sanitize_json(config)
        # config_json = json.dumps(safe_config, indent=4)
        # config_from_json = json.loads(config_json)

        filtered_results = []
        for result in self:
            config = sanitize_json(result.config)
            if dict_contains_dict(config, safe_config):
                filtered_results.append(result)

        return ResultSet(filtered_results)

    def filter(self, filter_fn: Callable[[Result], bool]) -> "ResultSet":
        """Create a copy of this `ResultSet` filtered by a function. The new `ResultSet`
        contains the results that satisfy the condition given by the filter function.

        Parameters
        ----------
        filter_fn : Callable[[Result], bool]
            Function to filter the results. If it returns True, the result will be
            included in the new `ResultSet`. The function receives a single parameter
            which is the `Result` object being processed. It must return a boolean value.

        Returns
        -------
        results : ResultSet
            A `ResultSet` that contains only the results that satisfy the condition.

        Raises
        ------
        TypeError
            If the `filter_fn` parameter is not a callable.
        """

        if not callable(filter_fn):
            raise TypeError(
                f"Expected callable, got {type(filter_fn).__name__}",
            )

        filtered_results = []
        for result in self:
            match = filter_fn(result)
            if not isinstance(match, bool):
                raise TypeError(
                    "The filter function must return a boolean value.",
                )
            if match:
                filtered_results.append(result)

        return ResultSet(filtered_results)

    def create_dataframe(
        self,
        config_columns: List[str] = [],
        filter_fn: Callable[[Result], bool] = lambda result: True,
        metrics_fn: Callable[
            [np.ndarray, np.ndarray], Dict[str, float]
        ] = lambda targets, predictions: {},
        include_train: bool = False,
        include_val: bool = False,
        best_params_columns: List[str] = [],
        n_jobs: int = -1,
        config_columns_prefix: str = "config_",
        best_params_columns_prefix: str = "best_",
        raise_errors: Literal["error", "warning", "ignore"] = "error",
    ):
        """Creates a `pandas.DataFrame` that contains all the results stored in this
        ResultSet. The DataFrame will contain the columns specified in config_columns,
        best_params_columns, and the metrics computed by metrics_fn. The metrics will be
        computed on the test set by default, but the train and validation metrics can be
        included using the include_train and include_val flags. If filter_fn parameter
        is provided, only the results that satisfy the condition will be included in the
        DataFrame.

        Parameters
        ----------
        config_columns : List[str], optional, default=[]
            List of columns from the config to include in the dataframe.
        filter_fn : Callable[[ResultData], bool], optional, default=lambda result: True
            Function to filter the results to include in the dataframe. If it returns
            True, the result row will be included. The function receives a single
            parameter which is the Result object being processed. It must return a
            boolean value.
        metrics_fn : Callable[[np.ndarray, np.ndarray], Dict[str, float]]
            Function that computes the metrics from the targets and predictions.
            It receives two numpy arrays, the targets and the predictions, and returns a
            dictionary where the key is the name of the metric and the value is the value
            of the metric. The shape of the numpy arrays depend on the kind of data that
            is stored within the ResultData. While any shape can be valid, the
            implementation of the metrics function must be coherent with the data stored
            in the ResultData.
        include_train : bool, optional, default=False
            Whether to include the metrics computed on the train set.
        include_val : bool, optional, default=False
            Whether to include the metrics computed on the validation set.
        best_params_columns : List[str], optional, default=[]
            List of columns from the best_params list to include in the dataframe.
        n_jobs : int, optional, default=-1
            The number of jobs to run in parallel (must be > 0). If -1, all CPUs are used.
            If 1 is given, no parallel computing code is used at all, which is useful for
            debugging. For n_jobs below -1, (n_cpus + 1 + n_jobs) are used. Thus for
            n_jobs = -2, all CPUs but one are used.
            joblib is used for parallel processing. If it is not installed, n_jobs
            will be set to 1 and a warning will be issued.
            The parallel backend is not specified within this function, so the user
            can set it using the `parallel_backend` of the joblib API.
        config_columns_prefix : str, optional, default = "config\\_"
            The prefix to add to the config columns. If '', no prefix is added. Note
            that using an empty prefix can result in column name conflicts.
        best_params_columns_prefix : str, optional, default = "best\\_"
            The prefix to add to the best_params columns. If empty string, no prefix is added. Note
            that using an empty prefix can result in column name conflicts.
        raise_errors : Literal["error", "warning", "ignore"], optional, default="error"
            Defines the behaviour when an error occurs during the creation of a row. See
            `remayn.result_set.utils.get_metric_columns_values` for more details.

        Returns
        -------
        df : pandas dataframe
            The dataframe with the results.

        Raises
        ------
        ValueError
            If n_jobs is set to 0.
        ValueError
            If raise_errors is not one of "error", "warning" or "ignore".
        """

        data = []
        columns = []

        if n_jobs == 0:
            raise ValueError(
                "n_jobs=0 is not supported. See n_jobs parameter in the documentation."
            )

        if raise_errors not in ["error", "warning", "ignore"]:
            raise ValueError(
                "raise_errors must be one of 'error', 'warning' or 'ignore'."
            )

        # Set n_jobs to 1 if joblib is not available
        if n_jobs != 1 and not importlib.util.find_spec("joblib"):
            warnings.warn(
                "joblib is not installed. n_jobs will be set to 1. To use parallel"
                " processing, install joblib. Set n_jobs=1 to run sequentially.",
                RuntimeWarning,
            )
            n_jobs = 1

        if n_jobs != 1:
            # Parallel processing
            import joblib

            # Populate the list of jobs that will be run in parallel
            joblist = [
                (
                    result,
                    config_columns,
                    metrics_fn,
                    include_train,
                    include_val,
                    best_params_columns,
                    config_columns_prefix,
                    best_params_columns_prefix,
                    raise_errors,
                )
                for result in self
                if filter_fn(result)
            ]

            with joblib.Parallel(n_jobs=n_jobs) as parallel:
                data = (
                    list(
                        parallel(
                            joblib.delayed(get_row_from_result)(*args)
                            for args in joblist
                        )
                    )
                    or []
                )
        else:
            for result in self:
                if filter_fn(result):
                    data.append(
                        get_row_from_result(
                            result,
                            config_columns,
                            metrics_fn,
                            include_train,
                            include_val,
                            best_params_columns,
                            config_columns_prefix,
                            best_params_columns_prefix,
                            raise_errors,
                        )
                    )

        # Save the columns of the row with the highest number of columns to avoid columns
        # sorting issues
        for row in data:
            if row:
                if len(row.keys()) > len(columns):
                    columns = row.keys()

        return pd.DataFrame(data)

    def add(self, result: Result):
        """Adds a `Result` object to the `ResultSet`. If the result already exists in
        the `ResultSet`, it will be replaced by the new one.

        Parameters
        ----------
        result : `Result`
            The `Result` object to add.

        Raises
        ------
        TypeError
            If the `result` parameter is not a `Result`.
        """

        if not isinstance(result, Result):
            raise TypeError(
                f"Expected Result, got {type(result).__name__}",
            )

        self.results_[str(sanitize_json(result.config))] = result

    def remove(self, key: Union[str, dict, Result]):
        """Removes a `Result` object identified by a key from the `ResultSet`.

        Parameters
        ----------
        key : Union[str, dict, Result]
            The key of the `Result` to remove. It can be given as:
            - a string: the config dict of the result transformed to string,
            - a dict: the config dict of the result,
            - a `Result` object.

        Raises
        ------
        TypeError
            If the `key` parameter is not a str, a dict or a `Result`.
        KeyError
            If the key is not found in the `ResultSet`.
        """

        if isinstance(key, str):
            del self.results_[key]
        elif isinstance(key, dict):
            del self.results_[str(sanitize_json(key))]
        elif isinstance(key, Result):
            del self.results_[str(sanitize_json(key.config))]
        else:
            raise TypeError(
                f"Expected str, dict or Result, got {type(key).__name__}",
            )

    def __iter__(self):
        """Returns an iterator to the `Result` objects contained in this `ResultSet`.

        Returns
        -------
        iter : Iterator[Result]
            An iterator to the `Result` objects in the `ResultSet`.
        """
        return iter(self.results_.values())

    def __len__(self):
        return len(self.results_)

    def __getitem__(self, key: Union[str, dict, Result]):
        """Gets an item from the `ResultSet` by its key.

        The key can be one of:
        - a string: the config dict of the result transformed to string,
        - a dict: the config dict of the result,
        - a `Result` object.

        If the key is a string, it is employed as the key to search in the
        `ResultSet`. If the key is a dict, it is json sanitized and transformed to
        string to be used as the key. If the key is a `Result` object, its config is
        sanitized and transformed to string to be used as the key.

        Parameters
        ----------
        key : Union[str, dict, Result]
            The key to search in the `ResultSet`.

        Returns
        -------
        result : `Result`
            The `Result` object that corresponds to the key.
        """

        if isinstance(key, str):
            return self.results_[key]
        elif isinstance(key, dict):
            return self.results_[str(sanitize_json(key))]
        elif isinstance(key, Result):
            return self.results_[str(sanitize_json(key.config))]
        else:
            raise TypeError(
                f"Expected str, dict or Result, got {type(key).__name__}",
            )

    def __str__(self):
        return (
            "ResultSet with"
            f" {len(self.results_)} result{'s' if len(self.results_) > 1 else ''}"
        )

    def __repr__(self):
        return self.__str__()

    def __contains__(self, result: Union[str, dict, Result]) -> bool:
        return self.contains(result)


class ResultFolder(ResultSet):
    """Stores a set of set of `Result` objects loaded from a directory.
    For each `Result` object, it stores the metadata of the experiment, such as the
    experiment info and the path where the result is stored. The predictions and targets
    are not loaded until needed.

    Attributes
    ----------
    base_path : Path
        The path where the results are stored.
    results_ : list of Result
        The list of `Result` objects stored in the `ResultSet`.
    """

    base_path: Path
    results_: Dict[str, Result]

    def __init__(self, base_path):
        self.base_path = Path(base_path)

        super().__init__(results=[])

        self.load()

    def load(self):
        """Loads the experiment info of all the results from the `base_path` directory.
        Only the metadata of the experiments is loaded, while the `ResultData` is not
        loaded until needed.

        It retrieves all the json files from `base_path`. Also, it checks that each
        json file has a corresponding pkl file. If the number of json files does not
        match the number of pkl files, a ValueError is raised.

        Raises
        ------
        ValueError
            If the the pickle file for a given json file is not found.

        Examples
        --------
        >>> from remayn.result_set import ResultFolder
        >>> rf = ResultFolder("./results")
        """

        self.results_ = {}

        json_files = list(self.base_path.rglob("*.json"))
        pkl_files = list(self.base_path.rglob("*.pkl"))

        # Check that all the json files have its corresponding pkl file
        if len(json_files) != len(pkl_files):
            for json_file in json_files:
                if json_file.with_suffix(".pkl") not in pkl_files:
                    raise FileNotFoundError(
                        f"Could not find pkl file for json file {json_file}",
                    )
            warnings.warn(
                f"Number of json files ({len(json_files)}) does not match"
                f" number of pkl files ({len(pkl_files)})",
            )

        for path in json_files:
            result = Result.load(self.base_path, path.stem)
            key = str(sanitize_json(result.config))
            self.results_[key] = result



def get_deep_item_from_dict(d, key, separator="."):
    """Access a nested item in a dictionary using a string separated by a given
    separator. Each separator indicates a level in the nested dictionary.
    For example, a.b.c will access the item d['a']['b']['c'].

    Parameters
    ----------
    d : dict
        The dictionary to access.
    key : str
        The key to access in the dictionary. It should be separated by the
        separator.
    separator : str, optional, default='.'
        The separator to use to split the key.

    Returns
    -------
    value : Any
        The value in the dictionary at the given key. None is returned if not found.
    """

    splitted = key.split(separator)
    value = d
    for k in splitted:
        if isinstance(value, dict) and k in value.keys():
            value = value[k]
        else:
            return None

    return value


def dict_contains_dict(d: dict, sub_d: dict) -> bool:
    """Check if a dictionary contains another dictionary.
    The dictionary can contain other dictionaries, that will be checked recursively.

    Parameters
    ----------
    d : dict
        The dictionary to check.
    sub_d : dict
        The dictionary to check if it is contained in `d`.

    Returns
    -------
    contains : bool
        True if `d` contains `sub_d`, False otherwise.
    """

    for key, value in sub_d.items():
        if key not in d:
            return False

        if isinstance(value, dict):
            if not dict_contains_dict(d[key], value):
                return False
        elif d[key] != value:
            return False

    return True



import json
import pickle
import shutil
import time
from hashlib import md5
from pathlib import Path
from typing import Optional, Union
from uuid import uuid4

import numpy as np

from ..utils import NonDefaultStrMethodError, sanitize_json
from .result_data import ResultData


class Result:
    """Represents the result of a experiment.
    It contains the path where the experiment ResultData is stored, along with the
    experiment information.
    The ResultData is only loaded when needed to save memory and time.

    Attributes
    ----------
    base_path: str
        Base path where all the experiments are stored.
    id: str
        Unique identifier of the experiment.
    config: dict
        Dictionary containing the parameters used in the experiment. All the elements
        in the dictionary must be JSON serializable. Objects contained in this dict
        should implement a custom __str__ method.
    data\\_: Optional[ResultData]
        Contains the `ResultData` when loaded or None if it was not loaded yet.
        This attribute should not be accessed directly. Use get_result() instead to
        make sure that the ResultData is properly loaded before accessing it.
    data_md5sum_: Optional[str]
        md5sum of the ResultData file. It is None if the file was not loaded yet or if
        creating a new Result that was not saved yet.
    created_at_: Optional[float]
        Timestamp when the experiment was created. It is None if the experiment was not
        saved yet.
    updated_at_: Optional[float]
        Timestamp when the experiment was last updated. It is None if the experiment was
        not saved yet.
    """

    base_path: Path
    id: str
    config: Optional[dict]
    data_: Optional[ResultData]
    data_md5sum_: Optional[str]
    created_at: Optional[float]
    updated_at: Optional[float]

    def __init__(
        self,
        base_path: Union[str, Path],
        id: Optional[str] = None,
        config: Optional[dict] = None,
    ):
        """Initializes the Result object.
        By default, it does not load the whole ResultData.

        Parameters
        ----------
        base_path: Union[str, Path]
            Base path where all the experiments are stored.
        id: Optional[str]
            Unique identifier of the experiment. Will be used for the file names. If
            None, a new unique identifier will be generated.
        config: Optional[dict]
            Dictionary containing the parameters used in the experiment.
        """

        self.base_path = Path(base_path)
        if id is None:
            self.id = str(uuid4())
        else:
            self.id = id
        self.config = config
        self.data_ = None
        self.data_md5sum_ = None
        self.created_at = None
        self.updated_at = None

    def get_data_path(self):
        """Gets the path where the ResultData is stored.

        Returns
        -------
        Path
            Path where the ResultData is stored.
        """

        return self.base_path / f"{self.id}.pkl"

    def get_info_path(self):
        """Gets the path where the experiment information is stored.

        Returns
        -------
        Path
            Path where the experiment information is stored.
        """

        return self.base_path / f"{self.id}.json"

    def __str__(self):
        s = f"Config: {json.dumps(sanitize_json(self.config), indent=4)}"
        if self.data_ is None:
            s += f"""
Results info path: {self.get_info_path()} (data not loaded)
"""
        else:
            s += f"""
Results info path: {self.get_info_path()}
Results data file: {self.get_data_path()}

Targets shape: {self.data_.targets.shape if self.data_.targets is not None else 'N/A'}
Predictions shape: {self.data_.predictions.shape if self.data_.predictions is not None else 'N/A'}
Train targets shape: {self.data_.train_targets.shape if self.data_.train_targets is not None else 'N/A'}
Train predictions shape: {self.data_.train_predictions.shape if self.data_.train_predictions is not None else 'N/A'}
Val targets shape: {self.data_.val_targets.shape if self.data_.val_targets is not None else 'N/A'}
Val predictions shape: {self.data_.val_predictions.shape if self.data_.val_predictions is not None else 'N/A'}

Time: {self.data_.time if self.data_.time is not None else 'N/A'}
Train history: {self.data_.train_history if self.data_.train_history is not None else 'N/A'}
Val history: {self.data_.val_history if self.data_.val_history is not None else 'N/A'}
Best params: {self.data_.best_params if self.data_.best_params is not None else 'N/A'}
"""
        return s

    def __repr__(self):
        return self.__str__()

    def __eq__(self, other: Union["Result", dict]):
        """Compares two `Result` objects considering only their config.
        It returns False if the type of the other object is not a `Result` or a dict.

        Parameters
        ----------
        other: Union[Result, dict]
            The other Result object or a dictionary containing the config of the
            other experiment.

        Returns
        -------
        bool
            True if the configs are equal and False otherwise.
        """

        if not isinstance(other, (Result, dict)):
            return False

        return self.compare_config(other)

    def compare_config(self, other: Union["Result", dict]) -> bool:
        """Compare the config of this Result with the config of `other` Result. It
        returns True if the configs are equal and False otherwise.

        Parameters
        ----------
        other: Union[Result, dict]
            The other Result object or a dictionary containing the config of the
            other experiment.

        Returns
        -------
        bool
            True if the configs are equal and False otherwise.
        """

        if isinstance(other, Result):
            return sanitize_json(self.config) == sanitize_json(other.config)
        elif isinstance(other, dict):
            return sanitize_json(self.config) == sanitize_json(other)
        else:
            raise TypeError(
                f"Expected a Result or a dict, but got {type(other)} instead."
            )

    def load_data(self, force=False):
        """Load the `ResultData` from the disk.
        This method reads the `ResultData` from the disk and stores it in the `data\\_`
        attribute. It also checks the integrity of the pickle file using the md5sum.
        This method is called automatically by get_result() when the `ResultData` is
        needed. However, you can call it manually to force the loading of the file.
        If the file was already loaded, this method does nothing, unless force=True is
        passed as an argument.

        Parameters
        ----------
        force: bool, optional, default=False
            If True, the file will be loaded even if it was already loaded.

        Raises
        ------
        FileNotFoundError
            If the `ResultData` does not exist.
        ValueError
            If the md5sum of the file does not match the one stored in the experiment
            information.
        """

        if self.data_ is not None and not force:
            return

        data_path = self.get_data_path()

        if not data_path.exists():
            raise FileNotFoundError(
                f"ResultData {data_path} does not exist."
                " The experiment is incomplete!"
            )

        with open(data_path, "rb") as f:
            content = f.read()

        md5sum = md5(content).hexdigest()

        if md5sum != self.get_md5sum():
            raise ValueError(
                f"ResultData {data_path} integrity check failed."
                " The file may have been modified after the experiment."
            )

        data = pickle.loads(content)
        self.data_ = data

    def get_md5sum(self):
        """Gets the md5sum of the ResultData file, which is stored in the experiment
        information.

        Returns
        -------
        str
            The md5sum of the ResultData file.
        """

        return self.data_md5sum_

    def get_data(self, force_reload=False):
        """Gets the `ResultData` of the experiment. If it was not loaded yet, it loads it
        from the disk. If the file was already loaded, it returns the stored object.
        This method should be used to access the `ResultData` instead of accessing the
        `data\\_` attribute directly.

        Parameters
        ----------
        force_reload: bool, optional, default=False
            If True, the `ResultData` will be reloaded even if it was already loaded.
            If False, it will only load the `ResultData` when it has not been loaded yet.

        Returns
        -------
        ResultData
            The `ResultData` object containing the results of the experiment. None if the
            `ResultData` is empty.
        """

        if self.data_ is None or force_reload:
            self.load_data(force=force_reload)

        return self.data_

    def set_data(self, data: ResultData):
        """Sets the ResultData of the experiment.
        This method should be used to set the ResultData instead of setting the `data\\_`
        attribute directly.

        Parameters
        ----------
        data: ResultData
            The ResultData object containing the results of the experiment.
        """

        self.data_ = data

    def get_experiment_info(self):
        """Gets all the experiment info as a dictionary, including experiment config,
        timestamps, md5sum of the ResultData file and the path where the ResultData is
        stored.

        Returns
        -------
        dict
            Dictionary containing all the experiment information.
        """

        return {
            "config": self.config,
            "data_md5sum": self.get_md5sum(),
            "created_at": self.created_at,
            "updated_at": self.updated_at,
        }

    def save(self):
        """Saves this `Result` to the disk.
        It saves the experiment information in the info_path file and the
        `ResultData` in the experiment_info_['results_path'] pickle file.
        If the files already exist, they will be overwritten.
        If the directory where the files should be saved does not exist, it will be
        created.

        Returns
        -------
        Result
            The `Result` object itself.
        """

        info_path = self.get_info_path()
        data_path = self.get_data_path()

        current_time = time.time()

        # Set time stamps
        if self.created_at is not None:
            self.updated_at = current_time
        else:
            self.created_at = current_time
            self.updated_at = current_time

        # Create directory
        self.base_path.mkdir(parents=True, exist_ok=True)

        # Save ResultData
        with open(data_path, "wb") as f:
            pickle.dump(self.data_, f)

        # Update md5sum based on new ResultData file
        with open(data_path, "rb") as f:
            self.data_md5sum_ = md5(f.read()).hexdigest()

        # Save experiment info
        experiment_info = self.get_experiment_info()
        safe_info = None
        try:
            safe_info = sanitize_json(experiment_info, accept_default_str=False)
        except NonDefaultStrMethodError:
            # Remove pickle file if an error occurs
            data_path.unlink()

            raise ValueError(
                "Experiment info contains some fields that are subject to change when"
                " the experiment is loaded from disk. Please, make sure that all the"
                " elements within the config are JSON serializable and show a"
                f" deterministic representation.\n{safe_info=}"
            )

        with open(info_path, "w") as f:
            json.dump(safe_info, f, indent=4)

        return self

    def delete(self, missing_ok=False):
        """Deletes the experiment information file (json) and the ResultData file
        (pickle) from the disk.

        Parameters
        ----------
        missing_ok: bool, optional, default=False
            If True, the method will not raise an error if the files do not exist.

        Raises
        ------
        FileNotFoundError
            If the experiment information file or the `ResultData` file does not exist
            and `missing_ok` is False.

        Returns
        -------
        bool
            True if the files were deleted successfully.
        """

        info_path = self.get_info_path()
        results_path = self.get_data_path()

        info_path.unlink(missing_ok=missing_ok)
        results_path.unlink(missing_ok=missing_ok)

        return True

    @staticmethod
    def load(base_path: Union[str, Path], id: str) -> "Result":
        """Loads a `Result` from the disk.
        It loads the experiment information and the `ResultData` from the disk and
        creates a new `Result` object with the loaded data.

        Parameters
        ----------
        base_path: Union[str, Path]
            Base path where all the experiments are stored.
        id: str
            Unique identifier of the experiment.

        Returns
        -------
        Result
            A new `Result` object with the loaded data.

        Raises
        ------
        FileNotFoundError
            If the experiment information file does not exist.
        ValueError
            If the experiment information file is not a valid json file.
            If the experiment information file does not contain the 'config' key.
            If the experiment information file does not contain the 'data_md5sum' key.

        Examples
        --------
        >>> from remayn.result import Result
        >>> result = Result.load("./results", "123")
        """

        result = Result(base_path=base_path, id=id)

        info_path = result.get_info_path()

        try:
            with open(info_path, "r") as f:
                info = json.load(f)
        except FileNotFoundError:
            raise FileNotFoundError(
                f"Experiment information file {info_path} does not exist."
            )

        if "config" not in info:
            raise ValueError(
                f"Experiment information file {info_path} does not contain the 'config'"
                " key. It is not a valid experiment information file."
            )
        result.config = info["config"]

        if "data_md5sum" not in info:
            raise ValueError(
                f"Experiment information file {info_path} does not contain the"
                " 'data_md5sum' key. It is not a valid experiment information file."
            )
        result.data_md5sum_ = info["data_md5sum"]

        result.created_at = info["created_at"] if "created_at" in info else None
        result.updated_at = info["updated_at"] if "updated_at" in info else None

        return result

    def copy_to(self, base_path: Union[str, Path], new_id: Optional[str] = None):
        """Copies the experiment information and the `ResultData` to a new directory.
        It creates a new `Result` object with the same data and saves it in the new
        directory. The new `Result` can have a new unique identifier.

        Parameters
        ----------
        base_path: Union[str, Path]
            Base path where the new experiment will be stored. Must be different from the
            current base path.
        new_id: Optional[str], optional, default=None
            Unique identifier of the new experiment. If None, the same identifier will be
            used.

        Returns
        -------
        Result
            A new `Result` object with the same data saved in the new directory.

        Examples
        --------
        >>> from remayn.result import Result
        >>> result = Result.load("./results", "123")
        >>> new_result = result.copy_to("./new_results")
        """

        base_path = Path(base_path)

        if self.base_path == base_path:
            raise ValueError(
                "The new base path must be different from the current base path."
            )

        if new_id is None:
            new_id = self.id

        base_path.mkdir(parents=True, exist_ok=True)

        source_info_path = self.get_info_path()
        source_data_path = self.get_data_path()

        dest_info_path = base_path / f"{new_id}.json"
        dest_data_path = base_path / f"{new_id}.pkl"

        if source_info_path.exists():
            shutil.copy(source_info_path, dest_info_path)

            if source_data_path.exists():
                shutil.copy(source_data_path, dest_data_path)

            new_result = Result.load(base_path, new_id)
        else:
            new_result = Result(base_path, new_id, self.config)
            new_result.set_data(self.get_data())
            new_result.save()

        return new_result


def make_result(
    base_path: Union[str, Path],
    config: dict,
    targets: np.ndarray,
    predictions: np.ndarray,
    train_targets: Optional[np.ndarray] = None,
    train_predictions: Optional[np.ndarray] = None,
    val_targets: Optional[np.ndarray] = None,
    val_predictions: Optional[np.ndarray] = None,
    time: Optional[float] = None,
    train_history: Optional[np.ndarray] = None,
    val_history: Optional[np.ndarray] = None,
    best_params: Optional[dict] = None,
    best_model: Optional[object] = None,
):
    """Helper function to create a `Result` object with the given data.
    It creates a `Result` object and the associated `ResultData`. The `Result` and
    the `ResultData` are not saved in the disk. You can call the `save()` method to
    save them.

    Parameters
    ----------
    base_path: Union[str, Path]
        Path of the main directory that will contain this `Result` and all the other
        results related to these experiments.
    config: dict
        Dictionary containing the parameters used in the experiment.
    targets: np.ndarray
        Array containing the targets of the experiment. Any shape can be used.
    predictions: np.ndarray
        Array containing the predictions of the experiment. Any shape can be used.
    train_targets: Optional[np.ndarray], optional, default=None
        Array containing the training targets of the experiment. Any shape can be used.
    train_predictions: Optional[np.ndarray], optional, default=None
        Array containing the training predictions of the experiment. Any shape can be
        used.
    val_targets: Optional[np.ndarray], optional, default=None
        Array containing the validation targets of the experiment. Any shape can be
        used.
    val_predictions: Optional[np.ndarray], optional, default=None
        Array containing the validation predictions of the experiment. Any shape can be
        used.
    time: Optional[float], optional, default=None
        Time spent to run the experiment.
    train_history: Optional[np.ndarray], optional, default=None
        Array containing the training history recorded during the training process. It
        should be a 1D array with the value of the error on each iteration.
    val_history: Optional[np.ndarray], optional, default=None
        Array containing the validation history recorded during the training process. It
        should be a 1D array with the value of the error on each iteration.
    best_params: Optional[dict], optional, default=None
        Dictionary containing the best parameters found during the experiment. It can be
        used in case that the experiment employs a cross-validation process. It can be
        left as None if the experiment does not use a cross-validation process or the
        cross-validation process is splitted in different experiments.
    best_model: Optional[object], optional, default=None
        Best model found during the experiment.

    Returns
    -------
    Result
        A new `Result` object with the given data. The `Result` is not saved in the
        disk. You can call the `save()` method to save it.

    Examples
    --------
    >>> import numpy as np
    >>> from remayn.result import make_result
    >>> targets = np.array([1, 2, 3])
    >>> predictions = np.array([1.1, 2.2, 3.3])
    >>> config = {"model": "linear_regression"}
    >>> result = make_result("results", config, targets, predictions)
    >>> result.save()
    """

    # Create a new Result (empty id to create a new one)
    result = Result(base_path=base_path, config=config)

    result.set_data(
        ResultData(
            targets=targets,
            predictions=predictions,
            train_targets=train_targets,
            train_predictions=train_predictions,
            val_targets=val_targets,
            val_predictions=val_predictions,
            time=time,
            train_history=train_history,
            val_history=val_history,
            best_params=best_params,
            best_model=best_model,
        )
    )

    return result



from typing import Optional, Union

import numpy as np

from ..utils import check_array


class ResultData:
    """Stores the results of a experiment.
    ResultData objects only contain data and are usually stored as pickle files.

    Attributes
    ----------
    targets: Union[np.ndarray, list]
        Numpy array of target values for the test set.
    predictions: Union[np.ndarray, list]
        Numpy array of predicted values for the test set.
    train_targets: Optional[Union[np.ndarray, list]], optional, default=None
        Numpy array of target values for the training set.
    train_predictions: Optional[Union[np.ndarray, list]], optional, default=None
        Numpy array of predicted values for the training set.
    val_targets: Optional[Union[np.ndarray, list]], optional, default=None
        Numpy array of target values for the validation set.
    val_predictions: Optional[Union[np.ndarray, list]], optional, default=None
        Numpy array of predicted values for the validation set.
    time: float, optional, default=None
        Time taken to run the experiment.
    train_history: Optional[Union[np.ndarray, list]], optional, default=None
        Training history of the model, represented as the value of the error on each
        iteration.
    val_history: Optional[Union[np.ndarray, list]], optional, default=None
        Validation history of the model, represented as the value of the error on each
        iteration.
    best_params: Optional[dict], optional, default=None
        Dictionary of the best parameters found during the experiment.
        Can be used in case that the experiment employes a cross-validation process.
    best_model: Optional[object], optional, default=None
        Best model found during the experiment.
    """

    def __init__(
        self,
        *,
        targets: Union[np.ndarray, list],
        predictions: Union[np.ndarray, list],
        train_targets: Optional[Union[np.ndarray, list]] = None,
        train_predictions: Optional[Union[np.ndarray, list]] = None,
        val_targets: Optional[Union[np.ndarray, list]] = None,
        val_predictions: Optional[Union[np.ndarray, list]] = None,
        time: Optional[float] = None,
        train_history: Optional[Union[np.ndarray, list]] = None,
        val_history: Optional[Union[np.ndarray, list]] = None,
        best_params: Optional[dict] = None,
        best_model: Optional[object] = None,
    ):
        targets = check_array(targets)
        predictions = check_array(predictions)
        train_targets = check_array(train_targets, allow_none=True)
        train_predictions = check_array(train_predictions, allow_none=True)
        val_targets = check_array(val_targets, allow_none=True)
        val_predictions = check_array(val_predictions, allow_none=True)
        train_history = check_array(train_history, allow_none=True)
        val_history = check_array(val_history, allow_none=True)

        if not isinstance(time, (float, int)) and time is not None:
            raise TypeError("time must be a float")

        if best_params is not None and not isinstance(best_params, dict):
            raise TypeError("best_params must be a dictionary")

        self.targets = targets
        self.predictions = predictions
        self.train_targets = train_targets
        self.train_predictions = train_predictions
        self.val_targets = val_targets
        self.val_predictions = val_predictions
        self.time = time
        self.train_history = train_history
        self.val_history = val_history
        self.best_params = best_params
        self.best_model = best_model

    def __eq__(self, other: "ResultData"):
        return (
            np.all(self.targets == other.targets)
            and np.all(self.predictions == other.predictions)
            and np.all(self.train_targets == other.train_targets)
            and np.all(self.train_predictions == other.train_predictions)
            and np.all(self.val_targets == other.val_targets)
            and np.all(self.val_predictions == other.val_predictions)
            and self.time == other.time
            and np.all(self.train_history == other.train_history)
            and np.all(self.val_history == other.val_history)
            and self.best_params == other.best_params
            and self.best_model == other.best_model
        )
